"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const core_1 = require("../core");
const http_proxy_middleware_1 = require("http-proxy-middleware");
const helpers_1 = require("../helpers");
const querystring = require("query-string");
/**
 * HTTP Reverse Proxy
 * ------------------
 * Attach a remote endpoint to an OpenAPI resource
 * The implementation of a HTTP Proxy plugin using the node-http-proxy package.
 *
 * See https://github.com/nodejitsu/node-http-proxy
 *
 * TODO: integrate with Vault to retrieve secrets and inject into egress request via header/params/etc.
 *
 */
class proxy {
    constructor() {
        this.name = "api.proxy";
        this.title = "Reverse proxy";
        this.proxyServer = null;
        /**
         * Create express middleware that forwards a request to a remote endpoint
         *
         * @param context
         * @param options
         * @param proxyPath
         * @returns {Function}
         * @constructor
         */
        this.proxy_middleware = function (context, _options, oper) {
            let options = this._target_options(_options);
            let mount_path = oper.resource;
            if (mount_path.endsWith("/*")) {
                mount_path = mount_path.substring(0, mount_path.length - 2);
            }
            context.log({
                code: "api:proxy:setup",
                message: oper.resource + " -> " + options.target,
                method: oper.actionId,
                mount_path: mount_path,
                proxyPath: oper.resource,
                options: options,
            });
            return (req, res, _next) => {
                // let params = _.extend({}, req.query, req.params);
                // let vars = _.extend({ resource: oper.resource, params: params });
                let params = helpers_1.APIs.getOperationContext(oper, req);
                try {
                    // interpolate vars in target path (template)
                    options.target = helpers_1.Vars.$(options.target, params);
                }
                catch (e) {
                    context.error({
                        code: "api:proxy:target:template:failed",
                        message: e.message,
                        resource: oper.resource,
                        params: params,
                    });
                }
                if (!options.onError) {
                    options.onError = (err) => {
                        console.log("onError: %j", err);
                        return helpers_1.Utils.sendError(context, res, {
                            code: "api:proxy:failed",
                            message: err.message || "unknown error",
                            error: err,
                            statusCode: 500,
                        });
                    };
                }
                // Work-around for body-parsers breaking proxy
                if (!options.onProxyReq)
                    options.onProxyReq = (proxyReq, req, _res) => {
                        const contentType = helpers_1.APIs.getContentType(proxyReq.getHeader("Content-Type"));
                        context.log({
                            code: "api:proxy:request",
                            message: "proxy",
                            request_uuid: req.a6s_uuid,
                            resource: oper.resource,
                            proxyPath: proxyReq.path,
                            orgPath: mount_path,
                            contentType: contentType,
                            // keys: _.keys(proxyReq),
                            options: options
                        });
                        // inject credentials into proxy request
                        this.reuseJWT(context, options, req, proxyReq);
                        if (!this.useCredentials(context, options, req, proxyReq)) {
                            throw new core_1.APIError(500, {
                                code: "api:proxy:credentials:missing",
                            });
                            return;
                        }
                        // remove mount-path from request
                        if (oper.resource.endsWith("/*") && proxyReq.path.startsWith(mount_path)) {
                            let proxymount_path = proxyReq.path;
                            proxyReq.path = proxyReq.path.substr(mount_path.length) || "/";
                            context.log({
                                code: "api:proxy:path:wildcard",
                                resource: oper.resource,
                                orgProxyPath: proxymount_path,
                                proxyPath: proxyReq.path,
                                proxyURL: proxyReq.url,
                                mount_path: mount_path,
                            });
                        }
                        try {
                            this.respond(contentType, proxyReq, req);
                        }
                        catch (err) {
                            return helpers_1.Utils.sendError(context, res, {
                                code: "api:proxy:response:failed",
                                message: err.message || "error",
                                error: err,
                                statusCode: 500,
                            });
                        }
                    };
                let proxy = http_proxy_middleware_1.createProxyMiddleware(options);
                return proxy(req, res, _next);
            };
        };
        /**
         * If enabled, copy the incoming JWT to the proxy request.
         * Otherwise, ensure the JWT is removed.
         *
         * @param context
         * @param options
         * @param req
         * @param proxyReq
         */
        this.reuseJWT = function (context, options, req, proxyReq) {
            if (options.reuseJWT) {
                let token = req.header("authorization");
                proxyReq.setHeader("authorization", token);
                context.log({
                    code: "api:proxy:jwt:reuse",
                    path: proxyReq.path
                });
            }
            else {
                // context.log({
                //     code: "api:proxy:jwt:redact",
                //     path: proxyReq.path
                // });
                proxyReq.removeHeader("authorization");
            }
        };
    }
    /**
     * plugin to proxy requests to a remote HTTP endpoint
     *
     * @param context
     * @param options
     * @returns {TargetEndpoint}
     */
    fn(operation, _options) {
        let context = operation.context;
        let options = _.extend({}, _options);
        return this.proxy_middleware(context, options, operation);
    }
    _target_options(options) {
        // set defaults - NOTE: different defaults to https://github.com/nodejitsu/node-http-proxy
        let target_options = _.extend({
            target: false,
            reuseJWT: false,
            selfHandleResponse: false,
            autoRewrite: true,
            xfwd: true,
            prependPath: true,
            ignorePath: false,
            changeOrigin: true,
            toProxy: false,
            logLevel: "warn",
        }, options);
        return target_options;
    }
    // inject secret from ENV into HEADER
    // see also: https://ec.haxx.se/http-auth.html
    useCredentials(_context, _options, _req, proxyReq) {
        let secretEnvKey = _options.secretEnvKey || "";
        let secretHeader = _options.secretHeader || "authorization";
        let secretHeaderType = _options.secretHeaderType || "raw";
        let secretValue = process.env[secretEnvKey] || false;
        // skip .. if not env-key specific
        if (!secretEnvKey)
            return true;
        let _msg = {
            code: "api:proxy:credentials",
            message: secretHeaderType + " " + secretEnvKey,
            headerType: secretHeaderType,
            secretHeader: secretHeader,
            secretEnvKey: secretEnvKey,
            secretFound: secretValue ? true : false,
        };
        _context.log(_msg);
        // fail if secret not found
        if (!secretValue) {
            return false;
        }
        // inject secret into header
        switch (secretHeaderType.toLowerCase()) {
            case "raw":
                proxyReq.setHeader(secretHeader, secretValue);
                break;
            case "basic":
                proxyReq.setHeader(secretHeader, "Basic " + secretValue);
                break;
            case "bearer":
                proxyReq.setHeader(secretHeader, "Bearer " + secretValue);
                break;
            default:
                proxyReq.removeHeader(secretHeader);
        }
        return true;
    }
    // re-write body based on content-type
    respond(contentType, proxyReq, req) {
        console.log("write proxy: %j -> %j -> %j", contentType, Buffer.byteLength(req.body), typeof req.body);
        if (!contentType || contentType === "application/json") {
            let body = JSON.stringify(req.body);
            proxyReq.setHeader("Content-Length", Buffer.byteLength(body));
            // console.log("write proxy: %j -> %j", Buffer.byteLength(body), body);
            proxyReq.write(body);
        }
        else if (contentType === "application/x-www-form-urlencoded") {
            let body = querystring.stringify(req.body);
            proxyReq.setHeader("Content-Length", Buffer.byteLength(body));
            proxyReq.write(body);
        }
        else {
            console.log("writeBody: %s -> %j -> %j", contentType, typeof req.body, req.body);
            proxyReq.write(Buffer.from(req.body, "UTF-8"));
        }
    }
}
exports.default = proxy;
//# sourceMappingURL=proxy.js.map