"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const middleware_1 = require("../middleware");
const fs = require("fs");
const axios_1 = require("axios");
const KeycloakCerts_1 = require("../helpers/KeycloakCerts");
/**
 * plugin has injects middleware 'after' every response is generated
 *
 * @param context
 * @param options
 * @returns {exports}
 */
class JWTPlugin {
    constructor() {
        this.ALLOWED_ALGOS = ["RS256"];
        this.name = "jwt";
        this.title = "Inject plugins 'jwt' middleware";
        this.secretOrCertificate = null;
        this.ignoreVerify = false;
        this.debug = false;
    }
    hasCertificate() {
        return !!this.secretOrCertificate;
    }
    getCertificate() {
        return this.secretOrCertificate;
    }
    install(_context, _options) {
        this.context = _context;
        let options = _options;
        this.debug = _options.debug || false;
        this.secretOrCertificate = options.secret || null;
        if (options.certificateFile) {
            this.secretOrCertificate = this._loadPublicKey(options.certificateFile);
        }
        this.ignoreVerify = options.verify === false;
        if (!this.ignoreVerify) {
            if (options.realmURL) {
                this._downloadPublicKey(options.realmURL);
            }
            if (options.keycloakURL) {
                this._downloadKeycloakCerts(options.keycloakURL);
            }
        }
        this.algo = options.algo || this.ALLOWED_ALGOS[0];
        _context.middleware.set(new middleware_1.jwt(this));
        this.context.log({
            code: "plugin:jwt:installed",
            algo: this.algo,
            verify: options.verify,
            ignoreVerify: this.ignoreVerify,
        });
    }
    _loadPublicKey(certificateFile) {
        const key = fs.readFileSync(certificateFile, "utf8");
        this.context.log({
            code: "plugin:jwt:cert:loaded",
            certificateFile: certificateFile,
            fileFound: !!key,
        });
        return key;
    }
    _downloadKeycloakCerts(url) {
        this.keycloakCerts = new KeycloakCerts_1.KeycloakCerts(this.context, url);
        this.context.log({ code: "plugin:jwt:keycloak:cert:loaded" });
        // console.log("keyCloakCerts: %j", this.keycloakCerts);
    }
    _downloadPublicKey(url) {
        axios_1.default({ url: url })
            .then((r) => {
            this.secretOrCertificate =
                "-----BEGIN RSA PUBLIC KEY-----\n" +
                    r.data.public_key +
                    "\n-----END RSA PUBLIC KEY-----";
            this.context.log({
                code: "plugin:jwt:cert:loaded",
                secretOrCertificate: this.secretOrCertificate,
            });
        })
            .catch((err) => {
            this.context.error({
                code: "plugin:jwt:cert:load:failed",
                message: err.message,
                url: url,
            });
        });
    }
}
exports.default = JWTPlugin;
//# sourceMappingURL=jwt.js.map